#include "meshcollider.h"
#include "particles.h"
#include "ray.h"


MeshCollider::MeshCollider(Mesh* mesh, double coef)
    : _mesh(mesh), _coef(coef)
{}

bool MeshCollider::collide(Particle * p, CollisionResponse& coll){
    if (p) {
        Eigen::Vector3f pos = Eigen::Vector3f(p->x.x(), p->x.y(), p->x.z());
        Eigen::Vector3f vel = Eigen::Vector3f(p->v.x(), p->v.y(), p->v.z());
        Ray r = Ray(pos, vel);
        Hit hit;

        if (_mesh->intersect(r, hit)) { // On a une intersection entre le mesh et le rayon (trajectoire de la particule)


            int f0 = _mesh->faceIndices()(hit.faceId() * 3);
            int f1 = _mesh->faceIndices()(hit.faceId() * 3 + 1);
            int f2 = _mesh->faceIndices()(hit.faceId() * 3 + 2);
            Eigen::Vector3f a = _mesh->positions().col(f1) - _mesh->positions().col(f0);
            Eigen::Vector3f b = _mesh->positions().col(f2) - _mesh->positions().col(f0);
            Eigen::Vector3f n = a.cross(b).normalized();

            Eigen::Vector3f x = _mesh->positions().col(f0);
            Eigen::Vector3f d = pos - x;


            double dotResult = d.dot(n);
            if (dotResult < 0.01 && n.dot(vel) < 0) { // Si on est a un epsilon pres du point d'impact
                // Calcul de la réponse

                CollisionResponse response = getResponse(p,
                                                         Eigen::Vector3d(x.x(), x.y(), x.z()),
                                                         Eigen::Vector3d(n.x(), n.y(), n.z()));

                coll.f = response.f;
                coll.v = response.v;

                return true;
            }
        }
    }
    return false;
}
